from __future__ import annotations

from typing import Literal

from prettyqt import core, multimedia
from prettyqt.qt import QtMultimedia
from prettyqt.utils import InvalidParamError, bidict


ERROR = bidict(
    none=QtMultimedia.QCamera.Error.NoError,
    camera=QtMultimedia.QCamera.Error.CameraError,
)

ErrorStr = Literal["none", "camera"]


EXPOSURE_MODE = bidict(
    auto=QtMultimedia.QCamera.ExposureMode.ExposureAuto,
    manual=QtMultimedia.QCamera.ExposureMode.ExposureManual,
    portrait=QtMultimedia.QCamera.ExposureMode.ExposurePortrait,
    night=QtMultimedia.QCamera.ExposureMode.ExposureNight,
    sports=QtMultimedia.QCamera.ExposureMode.ExposureSports,
    snow=QtMultimedia.QCamera.ExposureMode.ExposureSnow,
    beach=QtMultimedia.QCamera.ExposureMode.ExposureBeach,
    action=QtMultimedia.QCamera.ExposureMode.ExposureAction,
    landscape=QtMultimedia.QCamera.ExposureMode.ExposureLandscape,
    night_portrait=QtMultimedia.QCamera.ExposureMode.ExposureNightPortrait,
    threatre=QtMultimedia.QCamera.ExposureMode.ExposureTheatre,
    sunset=QtMultimedia.QCamera.ExposureMode.ExposureSunset,
    steady_photo=QtMultimedia.QCamera.ExposureMode.ExposureSteadyPhoto,
    fireworks=QtMultimedia.QCamera.ExposureMode.ExposureFireworks,
    party=QtMultimedia.QCamera.ExposureMode.ExposureParty,
    candle_light=QtMultimedia.QCamera.ExposureMode.ExposureCandlelight,
    barcode=QtMultimedia.QCamera.ExposureMode.ExposureBarcode,
)

ExposureModeStr = Literal[
    "auto",
    "manual",
    "portrait",
    "night",
    "sports",
    "snow",
    "beach",
    "action",
    "landscape",
    "night_portrait",
    "threatre",
    "sunset",
    "steady_photo",
    "fireworks",
    "party",
    "candle_light",
    "barcode",
]


FEATURE = bidict(
    color_temperature=QtMultimedia.QCamera.Feature.ColorTemperature,
    exposure_compensation=QtMultimedia.QCamera.Feature.ExposureCompensation,
    iso_sensitivity=QtMultimedia.QCamera.Feature.IsoSensitivity,
    manual_exposure_time=QtMultimedia.QCamera.Feature.ManualExposureTime,
    custom_focus_point=QtMultimedia.QCamera.Feature.CustomFocusPoint,
    focus_distance=QtMultimedia.QCamera.Feature.FocusDistance,
)

FeatureStr = Literal[
    "color_temperature",
    "exposure_compensation",
    "iso_sensitivity",
    "manual_exposure_time",
    "custom_focus_point",
    "focus_distance",
]

FLASH_MODE = bidict(
    off=QtMultimedia.QCamera.FlashMode.FlashOff,
    on=QtMultimedia.QCamera.FlashMode.FlashOn,
    auto=QtMultimedia.QCamera.FlashMode.FlashAuto,
)

FlashModeStr = Literal["off", "on", "auto"]


FOCUS_MODE = bidict(
    auto=QtMultimedia.QCamera.FocusMode.FocusModeAuto,
    auto_near=QtMultimedia.QCamera.FocusMode.FocusModeAutoNear,
    auto_far=QtMultimedia.QCamera.FocusMode.FocusModeAutoFar,
    focal=QtMultimedia.QCamera.FocusMode.FocusModeHyperfocal,
    infinity=QtMultimedia.QCamera.FocusMode.FocusModeInfinity,
    manual=QtMultimedia.QCamera.FocusMode.FocusModeManual,
)

FocusModeStr = Literal[
    "auto",
    "auto_near",
    "auto_far",
    "focal",
    "infinity",
    "manual",
]

TORCH_MODE = bidict(
    off=QtMultimedia.QCamera.TorchMode.TorchOff,
    on=QtMultimedia.QCamera.TorchMode.TorchOn,
    auto=QtMultimedia.QCamera.TorchMode.TorchAuto,
)

TorchModeStr = Literal["off", "on", "auto"]


WHITE_BALANCE = bidict(
    auto=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceAuto,
    manual=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceManual,
    sunlight=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceSunlight,
    cloudy=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceCloudy,
    shade=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceShade,
    tungsten=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceTungsten,
    fluorescent=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceFluorescent,
    flash=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceFlash,
    sunset=QtMultimedia.QCamera.WhiteBalanceMode.WhiteBalanceSunset,
)

WhiteBalanceStr = Literal[
    "auto",
    "manual",
    "sunlight",
    "cloudy",
    "shade",
    "tungsten",
    "fluorescent",
    "flash",
    "sunset",
]


class Camera(core.ObjectMixin, QtMultimedia.QCamera):
    def set_exposure_mode(self, mode: ExposureModeStr):
        """Set the exposure mode.

        Args:
            mode: exposure mode

        Raises:
            InvalidParamError: exposure mode does not exist
        """
        if mode not in EXPOSURE_MODE:
            raise InvalidParamError(mode, EXPOSURE_MODE)
        self.setExposureMode(EXPOSURE_MODE[mode])

    def get_exposure_mode(self) -> ExposureModeStr:
        """Return current exposure mode.

        Returns:
            exposure mode
        """
        return EXPOSURE_MODE.inverse[self.exposureMode()]

    def set_torch_mode(self, mode: TorchModeStr):
        """Set the torch mode.

        Args:
            mode: torch mode

        Raises:
            InvalidParamError: torch mode does not exist
        """
        if mode not in TORCH_MODE:
            raise InvalidParamError(mode, TORCH_MODE)
        self.setTorchMode(TORCH_MODE[mode])

    def get_torch_mode(self) -> TorchModeStr:
        """Return current torch mode.

        Returns:
            torch mode
        """
        return TORCH_MODE.inverse[self.torchMode()]

    def set_flash_mode(self, mode: FlashModeStr):
        """Set the flash mode.

        Args:
            mode: flash mode

        Raises:
            InvalidParamError: flash mode does not exist
        """
        if mode not in FLASH_MODE:
            raise InvalidParamError(mode, FLASH_MODE)
        self.setFlashMode(FLASH_MODE[mode])

    def get_flash_mode(self) -> FlashModeStr:
        """Return current flash mode.

        Returns:
            flash mode
        """
        return FLASH_MODE.inverse[self.flashMode()]

    def set_white_balance_mode(self, mode: WhiteBalanceStr):
        """Set the white balance mode.

        Args:
            mode: white balance mode

        Raises:
            InvalidParamError: white balance mode does not exist
        """
        if mode not in WHITE_BALANCE:
            raise InvalidParamError(mode, WHITE_BALANCE)
        self.setWhiteBalanceMode(WHITE_BALANCE[mode])

    def get_white_balance_mode(self) -> WhiteBalanceStr:
        """Return current white balance mode.

        Returns:
            white balance mode
        """
        return WHITE_BALANCE.inverse[self.whiteBalanceMode()]

    def get_supported_features(self) -> list[FeatureStr]:
        return FEATURE.get_list(self.supportedFeatures())

    def get_error(self) -> ErrorStr:
        """Return current error state.

        Returns:
            error state
        """
        return ERROR.inverse[self.error()]

    def get_camera_format(self) -> multimedia.CameraFormat:
        return multimedia.CameraFormat(self.cameraFormat())



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, bidict


POLICIES = bidict(
    none=QtGui.QActionGroup.ExclusionPolicy.None_,
    exclusive=QtGui.QActionGroup.ExclusionPolicy.Exclusive,
    exclusive_optional=QtGui.QActionGroup.ExclusionPolicy.ExclusiveOptional,
)

ExclusionPolicyStr = Literal["none", "exclusive", "exclusive_optional"]


class ActionGroup(core.ObjectMixin, QtGui.QActionGroup):
    def __init__(self, parent: QtCore.QObject | None = None):
        super().__init__(parent)  # type: ignore

    def __len__(self) -> int:
        return len(self.actions())

    def __getitem__(self, item: int) -> QtGui.QAction:
        return self.actions()[item]

    def set_exclusion_policy(self, policy: ExclusionPolicyStr | None):
        """Set exclusion policy to use.

        Args:
            policy: exclusion policy to use

        Raises:
            InvalidParamError: exclusion policy does not exist
        """
        if policy is None:
            policy = "none"
        if policy not in POLICIES:
            raise InvalidParamError(policy, POLICIES)
        self.setExclusionPolicy(POLICIES[policy])

    def get_exclusion_policy(self) -> ExclusionPolicyStr:
        """Return current exclusion policy.

        Returns:
            exclusion policy
        """
        return POLICIES.inverse[self.exclusionPolicy()]


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    action = ActionGroup()
    act = gui.Action()
    action.addAction(act)
    print(act in action)
    app.main_loop()



"""Custom_models module.

Contains custom models
"""

from .selectionmixin import SelectionMixin
from .regexmatchesmodel import RegexMatchesModel
from .columnitemmodel import ColumnItemModel, ColumnTableModel, ColumnItem
from .importlibdistributionmodel import ImportlibDistributionModel
from .listmixin import ListMixin
from .modelmixin import ModelMixin
from .nesteditem import NestedItem
from .nestedmodel import NestedModel
from .jsonmodel import JsonModel
from .fsspecmodel import FSSpecTreeModel
from .subsequencesortfilterproxymodel import SubsequenceSortFilterProxyModel

__all__ = [
    "SelectionMixin",
    "ImportlibDistributionModel",
    "RegexMatchesModel",
    "ColumnItemModel",
    "ColumnTableModel",
    "ListMixin",
    "ModelMixin",
    "ColumnItem",
    "NestedModel",
    "NestedItem",
    "JsonModel",
    "FSSpecTreeModel",
    "SubsequenceSortFilterProxyModel",
]



from __future__ import annotations

import contextlib
import functools
import os
import pathlib
import sys
from typing import TYPE_CHECKING, Literal

from deprecated import deprecated
import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import (
    InvalidParamError,
    colors,
    datatypes,
    get_repr,
    helpers,
    prettyprinter,
)


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


LayoutStr = Literal["horizontal", "vertical", "grid", "form", "stacked", "flow"]

QWIDGETSIZE_MAX = 16777215  # QtWidgets.QWIDGETSIZE_MAX


class WidgetMixin(core.ObjectMixin):
    box: QtWidgets.QLayout

    def __repr__(self) -> str:
        return get_repr(self)

    def resize(self, *size) -> None:
        if isinstance(size[0], tuple):
            super().resize(*size[0])
        else:
            super().resize(*size)

    def insertAction(
        self, position_or_action: int | QtGui.QAction, action: QtGui.QAction
    ):
        """Extend insertAction to also allow int index."""
        if isinstance(position_or_action, int):
            position_or_action = self.actions()[position_or_action]
        super().insertAction(position_or_action, action)

    def toggle_fullscreen(self):
        """Toggle between fullscreen and regular size."""
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def toggle_maximized(self):
        """Toggle between maximized and regular size."""
        if self.isMaximized():
            self.showNormal()
        else:
            self.showMaximized()

    def raise_to_top(self):
        if sys.platform.startswith("win"):
            import win32con
            from win32gui import SetWindowPos

            # set to always-on-top and disable it again. that way windows stays in front
            flag = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW
            win_id = self.winId()
            SetWindowPos(win_id, win32con.HWND_TOPMOST, 0, 0, 0, 0, flag)
            SetWindowPos(win_id, win32con.HWND_NOTOPMOST, 0, 0, 0, 0, flag)
        # state = (self.windowState() & ~Qt.WindowMinimized) | Qt.WindowActive
        # self.setWindowState(state)
        self.raise_()
        self.show()
        self.activateWindow()

    def set_icon(self, icon: datatypes.IconType) -> None:
        """Set the window icon.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon, color=colors.WINDOW_ICON_COLOR)
        self.setWindowIcon(icon)

    def get_icon(self) -> gui.Icon | None:
        icon = self.windowIcon()
        return None if icon.isNull() else gui.Icon(icon)

    @functools.singledispatchmethod
    def set_min_size(self, size: QtCore.QSize | tuple[int | None, int | None]) -> None:
        if isinstance(size, tuple):
            x = 0 if size[0] is None else size[0]
            y = 0 if size[1] is None else size[1]
            self.setMinimumSize(x, y)
        else:
            self.setMinimumSize(size)

    @set_min_size.register
    def _(self, x: int, y: int | None):
        self.set_min_size((x, y))

    @set_min_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_min_size((x, y))

    @functools.singledispatchmethod
    def set_max_size(self, size: QtCore.QSize | tuple[int | None, int | None]) -> None:
        if isinstance(size, tuple):
            x = QWIDGETSIZE_MAX if size[0] is None else size[0]
            y = QWIDGETSIZE_MAX if size[1] is None else size[1]
            self.setMaximumSize(x, y)
        else:
            self.setMaximumSize(size)

    @set_max_size.register
    def _(self, x: int, y: int | None):
        self.set_max_size((x, y))

    @set_max_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_max_size((x, y))

    def set_min_width(self, width: int | None) -> None:
        if width is None:
            width = 0
        self.setMinimumWidth(width)

    def set_max_width(self, width: int | None) -> None:
        if width is None:
            width = QWIDGETSIZE_MAX
        self.setMaximumWidth(width)

    def set_min_height(self, height: int | None) -> None:
        if height is None:
            height = 0
        self.setMinimumHeight(height)

    def set_max_height(self, height: int | None) -> None:
        if height is None:
            height = QWIDGETSIZE_MAX
        self.setMaximumHeight(height)

    def set_enabled(self, enabled: bool = True) -> None:
        self.setEnabled(enabled)

    def set_disabled(self) -> None:
        self.setEnabled(False)

    def set_title(self, title: str) -> None:
        self.setWindowTitle(title)

    def get_title(self) -> str:
        return self.windowTitle()

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, QtCore.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        self.setToolTip(tooltip)

    def set_font(
        self,
        font_name: str | None = None,
        font_size: int | None = None,
        weight: int | None = None,
        italic: bool = False,
    ) -> gui.Font:
        if font_size is None:
            font_size = -1
        if weight is None:
            weight = -1
        if font_name is None:
            font_name = self.font().family()
        font = gui.Font(font_name, font_size, weight, italic)
        self.setFont(font)
        return font

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_foreground_role(self) -> gui.palette.RoleStr:
        return gui.palette.ROLE.inverse[self.foregroundRole()]

    def set_foreground_role(self, role: gui.palette.RoleStr):
        if role not in gui.palette.ROLE:
            raise InvalidParamError(role, gui.palette.ROLE)
        self.setForegroundRole(gui.palette.ROLE[role])

    def get_background_role(self) -> gui.palette.RoleStr:
        return gui.palette.ROLE.inverse[self.backgroundRole()]

    def set_background_role(self, role: gui.palette.RoleStr):
        if role not in gui.palette.ROLE:
            raise InvalidParamError(role, gui.palette.ROLE)
        self.setBackgroundRole(gui.palette.ROLE[role])

    def set_window_flags(self, *flags: constants.WindowFlagStr, append: bool = False):
        for flag in flags:
            if flag not in constants.WINDOW_FLAGS:
                raise InvalidParamError(flag, constants.WINDOW_FLAGS)
        result = helpers.merge_flags(flags, constants.WINDOW_FLAGS)
        if append:
            result = result | self.windowFlags()
        self.setWindowFlags(result)

    def set_flags(
        self,
        minimize: bool | None = None,
        maximize: bool | None = None,
        close: bool | None = None,
        stay_on_top: bool | None = None,
        frameless: bool | None = None,
        window: bool | None = None,
        dialog: bool | None = None,
        tooltip: bool | None = None,
        tool: bool | None = None,
        customize: bool | None = None,
        window_title: bool | None = None,
    ) -> None:
        flags = {
            minimize: QtCore.Qt.WindowType.WindowMinimizeButtonHint,
            maximize: QtCore.Qt.WindowType.WindowMaximizeButtonHint,
            close: QtCore.Qt.WindowType.WindowCloseButtonHint,
            stay_on_top: QtCore.Qt.WindowType.WindowStaysOnTopHint,
            frameless: QtCore.Qt.WindowType.FramelessWindowHint,
            window: QtCore.Qt.WindowType.Window,
            dialog: QtCore.Qt.WindowType.Dialog,
            tooltip: QtCore.Qt.WindowType.ToolTip,
            tool: QtCore.Qt.WindowType.Tool,
            customize: QtCore.Qt.WindowType.CustomizeWindowHint,
            window_title: QtCore.Qt.WindowType.WindowTitleHint,
        }
        for k, v in flags.items():
            if k is not None:
                self.setWindowFlag(v, k)

    def set_attribute(
        self, attribute: constants.WidgetAttributeStr, state: bool = True
    ) -> None:
        if attribute not in constants.WIDGET_ATTRIBUTE:
            raise InvalidParamError(attribute, constants.WIDGET_ATTRIBUTE)
        self.setAttribute(constants.WIDGET_ATTRIBUTE[attribute], state)

    def set_attributes(self, **kwargs: bool) -> None:
        for attr, state in kwargs.items():
            if attr not in constants.WIDGET_ATTRIBUTE:
                raise InvalidParamError(attr, constants.WIDGET_ATTRIBUTE)
            self.setAttribute(constants.WIDGET_ATTRIBUTE[attr], state)  # type: ignore

    def set_modality(self, modality: constants.ModalityStr) -> None:
        """Set modality for the dialog.

        Args:
            modality: modality for the main window

        Raises:
            InvalidParamError: modality type does not exist
        """
        if modality not in constants.MODALITY:
            raise InvalidParamError(modality, constants.MODALITY)
        self.setWindowModality(constants.MODALITY[modality])

    def get_modality(self) -> constants.ModalityStr:
        """Get the current modality modes as a string.

        Returns:
            modality mode
        """
        return constants.MODALITY.inverse[self.windowModality()]

    def set_size_policy(
        self,
        horizontal: widgets.sizepolicy.SizePolicyStr | None = None,
        vertical: widgets.sizepolicy.SizePolicyStr | None = None,
    ) -> None:
        """Set the sizes policy.

        Args:
            horizontal: horizontal size policy
            vertical: vertical size policy
        """
        sp = self.get_size_policy()
        if horizontal is not None:
            sp.set_horizontal_policy(horizontal)
        if vertical is not None:
            sp.set_vertical_policy(vertical)
        self.setSizePolicy(sp)

    def get_size_policy(self) -> widgets.SizePolicy:
        qpol = self.sizePolicy()
        if isinstance(qpol, widgets.SizePolicy):
            return qpol
        return widgets.SizePolicy.clone(qpol)

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())

    def set_background_color(self, color: datatypes.ColorType) -> None:
        col_str = "" if color is None else colors.get_color(color).name()
        with self.edit_stylesheet() as ss:
            ss.backgroundColor.setValue(col_str)

    @contextlib.contextmanager
    def grab_mouse_events(
        self, cursor_shape: constants.CursorShapeStr | None = None
    ) -> Iterator[None]:
        if cursor_shape is not None:
            self.grabMouse(constants.CURSOR_SHAPE[cursor_shape])
        else:
            self.grabMouse()
        yield None
        self.releaseMouse()

    @contextlib.contextmanager
    def grab_keyboard_events(self) -> Iterator[None]:
        self.grabKeyboard()
        yield None
        self.releaseKeyboard()

    @contextlib.contextmanager
    def updates_off(self) -> Iterator[None]:
        updates = self.updatesEnabled()
        self.setUpdatesEnabled(False)
        yield None
        self.setUpdatesEnabled(updates)

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case None:
                ss = ""
            case os.PathLike():
                ss = pathlib.Path(ss).read_text()
            case qstylizer.style.StyleSheet():
                ss = str(ss)
        self.setStyleSheet(ss)

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        try:
            return qstylizer.parser.parse(self.styleSheet())
        except ValueError:
            return qstylizer.style.StyleSheet()

    @contextlib.contextmanager
    def edit_palette(self) -> Iterator[gui.Palette]:
        palette = gui.Palette(self.palette())
        yield palette
        self.setPalette(palette)

    @contextlib.contextmanager
    def edit_font(self) -> Iterator[gui.Font]:
        font = gui.Font(self.font())
        yield font
        self.setFont(font)

    @deprecated(reason="This context manager is deprecated, use edit_font instead.")
    @contextlib.contextmanager
    def current_font(self) -> Iterator[gui.Font]:
        with self.edit_font() as font:
            yield font

    @deprecated(reason="This method is deprecated, use set_context_menu_policy instead.")
    def set_contextmenu_policy(self, policy: constants.ContextPolicyStr) -> None:
        return self.set_context_menu_policy(policy)

    def set_context_menu_policy(self, policy: constants.ContextPolicyStr) -> None:
        """Set contextmenu policy for given item view.

        Args:
            policy: contextmenu policy to use

        Raises:
            InvalidParamError: policy does not exist
        """
        if policy not in constants.CONTEXT_POLICY:
            raise InvalidParamError(policy, constants.CONTEXT_POLICY)
        self.setContextMenuPolicy(constants.CONTEXT_POLICY[policy])

    @deprecated(reason="This method is deprecated, use get_context_menu_policy instead.")
    def get_contextmenu_policy(self) -> constants.ContextPolicyStr:
        """Return current contextmenu policy.

        Returns:
            contextmenu policy
        """
        return self.get_context_menu_policy()

    def get_context_menu_policy(self) -> constants.ContextPolicyStr:
        """Return current contextmenu policy.

        Returns:
            contextmenu policy
        """
        return constants.CONTEXT_POLICY.inverse[self.contextMenuPolicy()]

    def set_window_state(self, policy: constants.WindowStateStr) -> None:
        """Set window state for given item view.

        Args:
            policy: window state to use

        Raises:
            InvalidParamError: policy does not exist
        """
        if policy not in constants.WINDOW_STATES:
            raise InvalidParamError(policy, constants.WINDOW_STATES)
        self.setWindowState(constants.WINDOW_STATES[policy])

    def get_window_state(self) -> constants.WindowStateStr:
        """Return current window state.

        Returns:
            window state
        """
        return constants.WINDOW_STATES.inverse[self.windowState()]

    def set_custom_menu(self, method: Callable) -> None:
        self.set_context_menu_policy("custom")
        self.customContextMenuRequested.connect(method)

    def set_layout(
        self,
        layout: LayoutStr | QtWidgets.QLayout | None,
        margin: int | None = None,
        spacing: int | None = None,
    ):
        if layout is None:
            return
        match layout:
            case "horizontal" | "vertical":
                self.box = widgets.BoxLayout(layout)
            case "grid":
                self.box = widgets.GridLayout()
            case "form":
                self.box = widgets.FormLayout()
            case "stacked":
                self.box = widgets.StackedLayout()
            case "flow":
                from prettyqt import custom_widgets

                self.box = custom_widgets.FlowLayout()
            case QtWidgets.QLayout():
                self.box = layout
            case _:
                raise ValueError("Invalid Layout")
        self.setLayout(self.box)
        if margin is not None:
            self.box.set_margin(margin)
        if spacing is not None:
            self.box.setSpacing(spacing)

    def center(self, screen: int = 0) -> None:
        qr = self.frameGeometry()
        cp = gui.GuiApplication.screens()[screen].geometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def set_cursor(self, cursor: constants.CursorShapeStr | QtGui.QCursor) -> None:
        if isinstance(cursor, QtGui.QCursor):
            curs = cursor
        elif cursor in constants.CURSOR_SHAPE:
            curs = gui.Cursor(constants.CURSOR_SHAPE[cursor])
        else:
            raise InvalidParamError(cursor, constants.CURSOR_SHAPE)
        self.setCursor(curs)

    def set_focus_policy(self, policy: constants.FocusPolicyStr) -> None:
        """Set the way the widget accepts keyboard focus.

        Args:
            policy (str): Focus policy

        Raises:
            InvalidParamError: Description
        """
        if policy not in constants.FOCUS_POLICY:
            raise InvalidParamError(policy, constants.FOCUS_POLICY)
        self.setFocusPolicy(constants.FOCUS_POLICY[policy])

    def get_focus_policy(self) -> constants.FocusPolicyStr:
        """Return waay the widget accepts keyboard focus.

        Returns:
            str: Focus policy
        """
        return constants.FOCUS_POLICY.inverse[self.focusPolicy()]

    def set_font_size(self, size: int) -> None:
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)

    @deprecated(reason="This method is deprecated, use get_font_metrics instead.")
    def font_metrics(self) -> gui.FontMetrics:
        return self.get_font_metrics()

    def get_font_metrics(self) -> gui.FontMetrics:
        return gui.FontMetrics(self.fontMetrics())

    def get_font_info(self) -> gui.FontInfo:
        return gui.FontInfo(self.fontInfo())

    def set_margin(self, margin: int) -> None:
        self.setContentsMargins(margin, margin, margin, margin)

    def raise_dock(self) -> bool:
        w = self.find_parent(QtWidgets.QDockWidget)
        if w is None:
            return False
        w.setVisible(True)
        w.raise_()
        return True

    def set_mask(
        self,
        area: datatypes.RectType | QtGui.QRegion | None,
        typ: gui.region.RegionTypeStr = "rectangle",
    ):
        match area:
            case None:
                self.clearMask()
                return
            case tuple():
                area = gui.Region(*area, gui.region.REGION_TYPE[typ])
            case QtCore.QRect():
                area = gui.Region(area, gui.region.REGION_TYPE[typ])
        self.setMask(area)

    def set_window_file_path(self, path: datatypes.PathType):
        self.setWindowFilePath(os.fspath(path))

    def get_window_file_path(self) -> pathlib.Path | None:
        path = self.windowFilePath()
        return pathlib.Path(path) if path else None

    def get_image(self) -> QtGui.QPixmap:
        image = self.grab()
        with gui.Painter(image) as painter:
            painter.set_composition_mode("source_atop")
            for gl_widget in self.find_children(QtWidgets.QOpenGLWidget):
                d = gl_widget.mapToGlobal(core.Point()) - self.mapToGlobal(core.Point())
                painter.drawImage(d, gl_widget.grabFramebuffer())
        return image

    def get_screen(self) -> gui.Screen | None:
        window = self.window().windowHandle()
        return None if window is None else gui.Screen(window.screen())

    def delete_children(self):
        """Delete all children of the specified QObject."""
        if hasattr(self, "clear"):
            return self.clear()
        layout = self.layout()
        while layout.count():
            item = layout.takeAt(0)
            if widget := item.widget():
                widget.deleteLater()
            else:
                self.delete_children(item.layout())

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def set_style(self, style: str | QtWidgets.QStyle):
        if isinstance(style, str):
            style = QtWidgets.QStyleFactory.create(style)
        self.setStyle(style)


class Widget(WidgetMixin, prettyprinter.PrettyPrinter, QtWidgets.QWidget):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = Widget()
    widget.show()
    widget.set_min_size((400, 400))
    widget.set_max_size(None, 600)
    print(type(widget.get_screen()))
    app.main_loop()



from __future__ import annotations

import ctypes
from ctypes import POINTER, Structure, byref, c_int, cdll
from ctypes.wintypes import HWND, POINT, RECT, UINT
from typing import SupportsInt

import win32api
import win32con
import win32gui

from prettyqt.qt import QtWidgets


user32 = ctypes.windll.user32


class MARGINS(Structure):
    _fields_ = [
        ("cxLeftWidth", c_int),
        ("cxRightWidth", c_int),
        ("cyTopHeight", c_int),
        ("cyBottomHeight", c_int),
    ]


class PWINDOWPOS(Structure):
    _fields_ = [
        ("hWnd", HWND),
        ("hwndInsertAfter", HWND),
        ("x", c_int),
        ("y", c_int),
        ("cx", c_int),
        ("cy", c_int),
        ("flags", UINT),
    ]


class NCCALCSIZE_PARAMS(Structure):
    _fields_ = [("rgrc", RECT * 3), ("lppos", POINTER(PWINDOWPOS))]


LPNCCALCSIZE_PARAMS = POINTER(NCCALCSIZE_PARAMS)


def raise_to_top(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    # set to always-on-top and disable it again. that way windows stays in front
    flag = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW
    win32gui.SetWindowPos(h_wnd, win32con.HWND_TOPMOST, 0, 0, 0, 0, flag)
    win32gui.SetWindowPos(h_wnd, win32con.HWND_NOTOPMOST, 0, 0, 0, 0, flag)


def add_shadow_effect(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    margins = MARGINS(-1, -1, -1, -1)
    dwmapi = cdll.LoadLibrary("dwmapi")
    dwmapi.DwmExtendFrameIntoClientArea(h_wnd, byref(margins))


def add_window_animation(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    style = win32gui.GetWindowLong(h_wnd, win32con.GWL_STYLE)
    win32gui.SetWindowLong(
        h_wnd,
        win32con.GWL_STYLE,
        style
        | win32con.WS_MINIMIZEBOX
        | win32con.WS_MAXIMIZEBOX
        | win32con.WS_CAPTION
        | win32con.CS_DBLCLKS
        | win32con.WS_THICKFRAME,
        # | win32con.WS_MINIMIZEBOX,
        # | win32con.WS_MAXIMIZEBOX
        # | win32con.WS_CAPTION
        # | win32con.WS_THICKFRAME
        # | win32con.WS_POPUP
        # | win32con.WS_SYSMENU
    )


def set_window_nonresizable(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    style = win32gui.GetWindowLong(h_wnd, win32con.GWL_STYLE)
    style &= ~win32con.WS_SIZEBOX
    style &= ~win32con.WS_THICKFRAME
    style &= ~win32con.WS_MAXIMIZEBOX
    win32gui.SetWindowLong(h_wnd, win32con.GWL_STYLE, style)


def is_window_resizable(h_wnd: SupportsInt) -> bool:
    h_wnd = int(h_wnd)
    style = win32api.GetWindowLong(h_wnd, win32con.GWL_STYLE)
    return style & win32con.WS_SIZEBOX != 0


def is_maximized(h_wnd: SupportsInt) -> bool:
    h_wnd = int(h_wnd)
    window_placement = win32gui.GetWindowPlacement(h_wnd)
    return window_placement[1] == win32con.SW_MAXIMIZE if window_placement else False


def is_fullscreen(h_wnd: SupportsInt) -> bool:
    h_wnd = int(h_wnd)
    win_rect = win32gui.GetWindowRect(h_wnd)
    if not win_rect:
        return False

    monitor = win32api.MonitorFromWindow(h_wnd, win32con.MONITOR_DEFAULTTOPRIMARY)
    monitor_info = win32api.GetMonitorInfo(monitor)
    if not monitor_info:
        return False

    monitor_rect = monitor_info["Monitor"]
    return all(i == j for i, j in zip(win_rect, monitor_rect))


def start_system_move(h_wnd: SupportsInt) -> None:
    win32gui.ReleaseCapture()
    win32api.SendMessage(
        int(h_wnd), win32con.WM_SYSCOMMAND, win32con.SC_MOVE | win32con.HTCAPTION, 0
    )


def _native_event(widget: QtWidgets.QWidget, message: SupportsInt) -> tuple[bool, int]:
    pt = POINT()
    ctypes.windll.user32.GetCursorPos(ctypes.byref(pt))
    r = widget.devicePixelRatioF()
    x = pt.x / r - widget.x()
    y = pt.y / r - widget.y()
    is_maximize_button = widget.childAt(x, y) is widget.title_bar.maximize_button
    msg = ctypes.wintypes.MSG.from_address(int(message))
    dpi = user32.GetDpiForWindow(msg.hWnd)
    sys_metrics = user32.GetSystemMetricsForDpi(92, dpi)
    border_width = user32.GetSystemMetricsForDpi(win32con.SM_CXFRAME, dpi) + sys_metrics
    border_height = user32.GetSystemMetricsForDpi(win32con.SM_CYFRAME, dpi) + sys_metrics
    match msg.message:
        case win32con.WM_NCHITTEST:
            if widget.is_resizable():
                w, h = widget.width(), widget.height()
                lx = x < border_width
                rx = x > w - 2 * border_width
                ty = y < border_height
                by = y > h - border_height

                if lx and ty:
                    return True, win32con.HTTOPLEFT
                if rx and by:
                    return True, win32con.HTBOTTOMRIGHT
                if rx and ty:
                    return True, win32con.HTTOPRIGHT
                if lx and by:
                    return True, win32con.HTBOTTOMLEFT
                if ty:
                    return True, win32con.HTTOP
                if by:
                    return True, win32con.HTBOTTOM
                if lx:
                    return True, win32con.HTLEFT
                if rx:
                    return True, win32con.HTRIGHT

            if is_maximize_button:
                widget.title_bar.maximize_button.set_state("hover")
                return True, win32con.HTMAXBUTTON
            buttons = widget.title_bar.findChildren(QtWidgets.QPushButton)
            if (
                widget.childAt(x, y) not in buttons
                and border_height < y < widget.title_bar.height()
            ):
                return True, win32con.HTCAPTION

        case 0x2A2 | win32con.WM_MOUSELEAVE:
            widget.title_bar.maximize_button.set_state("normal")
        case win32con.WM_NCLBUTTONDOWN | win32con.WM_NCLBUTTONDBLCLK:
            if is_maximize_button:
                widget.title_bar.on_maximize_button_clicked()
                return True, 0
        case win32con.WM_NCLBUTTONUP | win32con.WM_NCRBUTTONUP:
            if is_maximize_button:
                widget.title_bar.on_maximize_button_clicked()
                return None
        case win32con.WM_NCCALCSIZE:
            rect = ctypes.cast(msg.lParam, LPNCCALCSIZE_PARAMS).contents.rgrc[0]

            is_max = is_maximized(msg.hWnd)
            is_full = is_fullscreen(msg.hWnd)

            # adjust the size of client rect
            if is_max and not is_full:
                rect.top += border_height
                rect.left += border_width
                rect.right -= border_width
                rect.bottom -= border_height

            return True, win32con.WVR_REDRAW



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


class SignalMapper(core.ObjectMixin, QtCore.QSignalMapper):
    def __getitem__(self, index: int | str | QtCore.QObject) -> QtCore.QObject:
        return self.mapping(index)

    def __setitem__(self, index: QtCore.QObject, value: int | str | QtCore.QObject):
        self.setMapping(index, value)


if __name__ == "__main__":
    mapper = SignalMapper()



from __future__ import annotations

from prettyqt import constants, core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError


class Drag(core.ObjectMixin, QtGui.QDrag):
    def get_pixmap(self) -> gui.Pixmap:
        return gui.Pixmap(self.pixmap())

    def get_default_action(self) -> constants.DropActionStr:
        return constants.DROP_ACTION.inverse[self.defaultAction()]

    def get_drag_cursor(self, action: constants.DropActionStr) -> gui.Pixmap:
        if action not in constants.DROP_ACTION:
            raise InvalidParamError(action, constants.DROP_ACTION)
        px = self.dragCursor(constants.DROP_ACTION[action])
        return gui.Pixmap(px)

    def set_drag_cursor(self, cursor: QtGui.QPixmap, action: constants.DropActionStr):
        self.setDragCursor(cursor, constants.DROP_ACTION[action])

    def get_supported_actions(self) -> list[constants.DropActionStr]:
        return constants.DROP_ACTION.get_list(self.supportedActions())

    def main_loop(
        self,
        supported_actions: list[constants.DropActionStr] | None = None,
        default_drop_action: constants.DropActionStr | None = None,
    ):
        supported_actions = supported_actions or [constants.DROP_ACTION["move"]]
        flag = QtCore.Qt.DropAction(0)
        for i in supported_actions:
            flag |= constants.DROP_ACTION[i]
        default_action = constants.DROP_ACTION[default_drop_action]
        self.exec(flag, default_action)


if __name__ == "__main__":
    obj = core.Object()
    drag = Drag(obj)
    drag.get_pixmap()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


# TYPES = {
#     bool: 1,
#     int: 2,
#     str: 10,
#     float: 38,
#     QtGui.QColor: 67,
#     QtGui.QCursor: 74,
#     QtCore.QDate: 14,
#     QtCore.QSize: 21,
#     QtCore.QTime: 15,
#     list: 9,
#     QtGui.QPolygon: 71,
#     QtGui.QPolygonF: 86,
#     QtGui.QColor: 67,
#     QtGui.QColorSpace: 87,
#     QtCore.QSizeF: 22,
#     QtCore.QRectF: 20,
#     QtCore.QLine: 23,
#     QtGui.QTextLength: 77,
#     dict: 8,
#     QtGui.QIcon: 69,
#     QtGui.QPen: 76,
#     QtCore.QLineF: 24,
#     QtGui.QTextFormat: 78,
#     QtCore.QRect: 19,
#     QtCore.QPoint: 25,
#     QtCore.QUrl: 17,
#     QtCore.QRegularExpression: 44,
#     QtCore.QDateTime: 16,
#     QtCore.QPointF: 26,
#     QtGui.QPalette: 68,
#     QtGui.QFont: 64,
#     QtGui.QBrush: 66,
#     QtGui.QRegion: 72,
#     QtGui.QImage: 70,
#     QtGui.QKeySequence: 75,
#     QtWidgets.QSizePolicy: 121,
#     QtGui.QPixmap: 65,
#     QtCore.QLocale: 18,
#     QtGui.QBitmap: 73,
#     QtGui.QMatrix4x4: 81,
#     QtGui.QVector2D: 82,
#     QtGui.QVector3D: 83,
#     QtGui.QVector4D: 84,
#     QtGui.QQuaternion: 85,
#     QtCore.QEasingCurve: 29,
#     QtCore.QJsonValue: 45,
#     QtCore.QJsonDocument: 48,
#     QtCore.QModelIndex: 42,
#     QtCore.QPersistentModelIndex: 50,
#     QtCore.QUuid: 30,
#     "user": 1024,
# }


class MetaObject:
    def __init__(self, metaobject: QtCore.QMetaObject):
        self.item = metaobject

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_super_class(self) -> MetaObject | None:
        return MetaObject(klass) if (klass := self.superClass()) is not None else None

    def get_class_info(self, include_super: bool = True) -> dict[str, str]:
        start = 0 if include_super else self.item.classInfoOffset() - 1
        count = self.item.classInfoCount()
        classinfos = [self.item.classInfo(i) for i in range(start, count)]
        return {i.name(): i.value() for i in classinfos}

    def get_method(self, index: int | str) -> core.MetaMethod:
        if isinstance(index, int):
            return core.MetaMethod(self.item.method(index))
        for method in self.get_methods():
            if method.get_name() == index:
                return method
        raise KeyError(index)

    def get_enum(self, index: int | str) -> core.MetaEnum:
        if isinstance(index, int):
            return core.MetaEnum(self.item.enumerator(index))
        for enumerator in self.get_enums():
            if enumerator.get_name() == index:
                return enumerator
        raise KeyError(index)

    def get_property(self, index: int | str) -> core.MetaProperty:
        if isinstance(index, int):
            return core.MetaProperty(self.item.property(index))
        for prop in self.get_properties():
            if prop.get_name() == index:
                return prop
        raise KeyError(index)

    def get_constructor(self, index: int | str) -> core.MetaProperty:
        if isinstance(index, int):
            return core.MetaMethod(self.item.constructor(index))
        for prop in self.get_constructors():
            if prop.get_name() == index:
                return prop
        raise KeyError(index)

    def get_methods(
        self,
        include_super: bool = True,
        type_filter: core.metamethod.MethodTypeStr | None = None,
    ) -> list[core.MetaMethod]:
        start = 0 if include_super else self.item.methodOffset() - 1
        methods = [self.get_method(i) for i in range(start, self.item.methodCount())]
        if type_filter is None:
            return methods
        else:
            return [i for i in methods if i.get_method_type() == type_filter]

    def get_enums(self, include_super: bool = True) -> list[core.MetaEnum]:
        start = 0 if include_super else self.item.enumeratorOffset() - 1
        return [self.get_enum(i) for i in range(start, self.item.enumeratorCount())]

    def get_constructors(self) -> list[core.MetaMethod]:
        count = self.item.constructorCount()
        return [core.MetaMethod(self.item.constructor(i)) for i in range(count)]

    def get_properties(self, include_super: bool = True) -> list[core.MetaProperty]:
        start = 0 if include_super else self.item.propertyOffset() - 1
        count = self.item.propertyCount()
        return [core.MetaProperty(self.item.property(i)) for i in range(start, count)]

    def get_signals(self, include_super: bool = True) -> list[core.MetaMethod]:
        return [
            i
            for i in self.get_methods(include_super=include_super)
            if i.get_method_type() == "signal"
        ]

    def get_meta_type(self) -> core.MetaType:
        return core.MetaType(self.metaType().id())


if __name__ == "__main__":
    from prettyqt import gui

    metaobj = gui.StyleHints.get_metaobject()
    print(metaobj.get_meta_type())



from __future__ import annotations

from collections.abc import Iterable
import datetime
import enum
import itertools
import os
import pathlib
from typing import TypedDict

from prettyqt import constants, core, custom_models, widgets
from prettyqt.qt import QtCore
from prettyqt.utils import treeitem


class FolderInfo(TypedDict):
    name: str
    size: int
    type: str
    created: float
    islink: bool  # symbolic link
    mode: int  # Inode protection mode.
    uid: int  # user id of owner
    gid: int  # group id of owner
    mtime: float
    ino: int  # Inode number.
    nlink: int  # Number of links to the inode.


_icon_provider = widgets.FileIconProvider()


VALUE_LETTERS = [(4, "r"), (2, "w"), (1, "x")]


def octal_to_string(octal: int) -> str:
    result = ""
    # Iterate over each of the digits in octal
    for digit, (value, letter) in itertools.product(
        [int(n) for n in str(octal)], VALUE_LETTERS
    ):
        if digit >= value:
            result += letter
            digit -= value
        else:
            result += "-"
    return result


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def get_filename(path):
    return pathlib.Path(path).name if path else ""


ATTR_MODEL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="The name of the object.",
    label=lambda x: get_filename(x.obj["name"]),
    set_edit=lambda x, value: x.set_name(value),
    decoration=lambda x: _icon_provider.get_icon(core.FileInfo(x.obj["name"])),
    user_data={
        constants.USER_ROLE + 1: lambda x: x.obj["name"],
        constants.USER_ROLE + 2: lambda x: get_filename(x.obj["name"]),
    },
)


ATTR_MODEL_PATH = custom_models.ColumnItem(
    name="Path",
    doc="A path",
    label=lambda x: x.obj["name"] or "",
)

ATTR_MODEL_SIZE = custom_models.ColumnItem(
    name="Size",
    doc="Item size.",
    label=lambda x: sizeof_fmt(x.obj["size"]) if x.obj["size"] > 0 else "",
    sort_value=lambda x: x.obj["size"],
)

ATTR_MODEL_TYPE = custom_models.ColumnItem(
    name="Type",
    doc="Item type.",
    label=lambda x: x.obj["type"] or "",
)

ATTR_MODEL_CREATED = custom_models.ColumnItem(
    name="Created",
    doc="Date created.",
    label=lambda x: datetime.datetime.fromtimestamp(x.obj["created"])
    if x.obj.get("created")
    else "",
)

ATTR_MODEL_MODIFIED = custom_models.ColumnItem(
    name="Modified",
    doc="Date modified.",
    label=lambda x: datetime.datetime.fromtimestamp(x.obj["mtime"])
    if x.obj.get("mtime")
    else "",
)

ATTR_MODEL_PERMISSIONS = custom_models.ColumnItem(
    name="Permissions",
    doc="File permissions.",
    label=lambda x: oct(int(x.obj["mode"]))[-4:] if x.obj.get("mode") else "",
)

ATTR_MODEL_IS_LINK = custom_models.ColumnItem(
    name="Link",
    label="",
    doc="Date created.",
    checkstate=lambda x: x.obj.get("islink") or False,
)


COLUMNS = [
    ATTR_MODEL_NAME,
    ATTR_MODEL_PATH,
    ATTR_MODEL_TYPE,
    ATTR_MODEL_MODIFIED,
    ATTR_MODEL_SIZE,
    ATTR_MODEL_CREATED,
    ATTR_MODEL_PERMISSIONS,
    ATTR_MODEL_IS_LINK,
]


class FSSpecTreeModel(
    widgets.filesystemmodel.FileSystemModelMixin, custom_models.ColumnItemModel
):
    """Model that provides an interface to an objectree that is build of tree items."""

    directoryLoaded = core.Signal(str)
    fileRenamed = core.Signal(str, str, str)
    rootPathChanged = core.Signal(str)

    @core.Enum
    class Roles(enum.IntEnum):
        """Role enum."""

        FileIconRole = constants.DECORATION_ROLE
        FilePathRole = constants.USER_ROLE + 1
        FileNameRole = constants.USER_ROLE + 2
        FilePermissions = constants.USER_ROLE + 3

    @core.Enum
    class Option(enum.IntEnum):
        """Role enum."""

        DontWatchForChanges = 1
        DontResolveSymlinks = 2
        DontUseCustomDirectoryIcons = 4

    def __init__(
        self,
        fs,
        obj: FolderInfo,
        show_root: bool = False,
        parent: QtCore.QObject | None = None,
    ):
        super().__init__(attr_cols=COLUMNS, parent=parent)
        self._show_root = show_root
        self.fs = fs
        self.root_marker = fs.root_marker
        self.set_root_item(obj)

    def _fetch_object_children(self, obj: treeitem.TreeItem) -> list[treeitem.TreeItem]:
        """Fetch the children of a Python object.

        Returns: list of TreeItems
        """
        items = [
            treeitem.TreeItem(obj=i, parent=obj)
            for i in self.fs.glob(f"{obj.obj['name']}/*/", detail=True).values()
        ]
        # not sure if this should be emitted later?
        self.directoryLoaded.emit(obj.obj["name"])
        return items

    def get_protocol_path(self, index: QtCore.QModelIndex) -> str:
        protocol = self.fs.protocol
        path = index.data(self.Roles.FilePathRole)
        return f"{protocol}://{path}"

    def fileIcon(self, index):
        tree_item = index.internalPointer()
        if tree_item is None:
            return None
        return _icon_provider.get_icon(core.FileInfo(tree_item.obj["name"]))

    def fileInfo(self, index):
        tree_item = index.internalPointer()
        return None if tree_item is None else core.FileInfo(tree_item.obj["name"])

    def fileName(self, index) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else pathlib.Path(tree_item.obj["name"]).name

    def filePath(self, index) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else tree_item.obj["name"]

    def isDir(self, index):
        tree_item = index.internalPointer()
        return False if tree_item is None else tree_item.obj["type"] == "directory"

    def mkdir(self, index, name: str):
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        new_folder = pathlib.Path(path) / name
        self.fs.mkdir(str(new_folder))

    def size(self, index) -> int:
        tree_item = index.internalPointer()
        return 0 if tree_item is None else tree_item.obj["size"]

    def lastModified(self, index) -> int:
        tree_item = index.internalPointer()
        return (
            ""
            if tree_item is None
            else str(datetime.datetime.fromtimestamp(tree_item.obj["mtime"]))
        )

    def permissions(self, index) -> QtCore.QFileDevice.Permission:
        tree_item = index.internalPointer()
        flag = QtCore.QFileDevice.Permission(0)
        if tree_item is None:
            return flag
        val = oct(int(tree_item.obj["mode"]))[-4:]
        for i in core.filedevice.PERMISSIONS.get_list(int(val, 8)):
            flag |= core.filedevice.PERMISSIONS[i]
        return flag

    def remove(self, index) -> bool:
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        self.fs.rm_file(path)
        return self.fs.exists(path)

    def rmdir(self, index) -> bool:
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        self.fs.rmdir(path)
        return self.fs.exists(path)

    def setRootPath(self, path: os.PathLike) -> core.ModelIndex:
        path = os.fspath(path)
        self.set_root_item({"name": path, "size": 0, "type": "directory"})

        self.rootPathChanged.emit(path)

    def rootDirectory(self) -> core.Dir:
        return core.Dir(self.rootPath())

    def rootPath(self) -> str:
        return self._root_item["name"]

    def rename(self, index, value: str):
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        old_name = pathlib.Path(path).name
        parent_folder = pathlib.Path(path).parent
        target = str(parent_folder / value)
        self.fs.mv(path, target)
        exists = self.fs.exists(target)
        if exists:
            self.fileRenamed.emit(parent_folder, old_name, value)
        return exists

    def type(self, index) -> str:
        tree_item = index.internalPointer()
        return tree_item.obj["type"]

    def mimeTypes(self):
        return ["text/uri-list"]

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        paths = [i for idx in indexes if (i := idx.data(self.Roles.FilePathRole))]
        urls = [core.Url(i) for i in paths]
        mime_data.setUrls(urls)
        return mime_data

    def supportedDropActions(self):
        return (
            QtCore.Qt.DropAction.MoveAction
            | QtCore.Qt.DropAction.CopyAction
            | QtCore.Qt.DropAction.LinkAction
        )

    def flags(self, index):
        flags = super().flags(index)
        return flags | constants.DROP_ENABLED | constants.DRAG_ENABLED

    # def dragEnterEvent(self, event):
    #     event.accept() if event.mimeData().hasUrls() else super().dragEnterEvent(event)

    def hasChildren(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return False
        if self._show_root and self.tree_item(parent) == self._root_item:
            return True
        return self.tree_item(parent).obj["type"] == "directory"

    def removeRows(self, row: int, count: int, parent):
        print(row, count, parent.data())
        # end_row = row + count - 1
        # with self.remove_rows(row, end_row, parent):
        #     for i in range(end_row, row - 1, -1):
        #         self.items.pop(i)
        return True

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)

    def add_items(
        self,
        items: Iterable,
        position: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        if position is None:
            position = len(self.items)
        items = list(items)
        # with self.insert_rows(position, position + len(items) - 1, parent):
        #     for i in range(len(items)):
        #         self.items.insert(i + position, items[i])
        #         pass
        #     self.items.extend(items)
        return items

    def canDropMimeData(
        self,
        mime_data: QtCore.QMimeData,
        action: QtCore.Qt.DropAction,
        row: int,
        column: int,
        parent_index: QtCore.QModelIndex,
    ):
        return column == 0 and mime_data.hasFormat("text/uri-list")

    def dropMimeData(
        self,
        mime_data: QtCore.QMimeData,
        action: QtCore.Qt.DropAction,
        row: int,
        column: int,
        parent_index: QtCore.QModelIndex,
    ):
        if not self.canDropMimeData(mime_data, action, row, column, parent_index):
            return False
        print(mime_data.urls(), action, row, column, parent_index.data())
        urls = [core.Url(i) for i in mime_data.urls()]
        if not urls:
            return False
        if action == QtCore.Qt.DropAction.MoveAction:
            with self.change_layout():
                for i in sorted(urls, reverse=True):
                    # self.fs.mv(i, )
                    print("move", i, "to", parent_index.data(self.Roles.FilePathRole))
        elif action == QtCore.Qt.DropAction.CopyAction:
            pass
        elif action == QtCore.Qt.DropAction.LinkAction:
            pass

        column = min(column, 0)
        if row != -1:
            begin_row = row
        elif parent_index.isValid():
            begin_row = self.rowCount(parent_index) + 1
        else:
            begin_row = self.rowCount() + 1
        print(begin_row)
        return True

        # if action == Qt.CopyAction:
        #     # Behavior if parent is a text item
        #     # For example, we select a text and do: CTRL+C CTRL+V
        #     if parent.isValid() and not parent.internalPointer().isFolder():
        #         # We insert copy in parent folder, just below
        #         begin_row = parent.row() + 1
        #         parent = parent.parent()

        #     if parent.isValid() and parent.internalPointer().isFolder():
        #         pass
        # if not items:
        #     return False

        # if action == Qt.CopyAction:
        #     pass

        # r = self.insertItems(items, begin_row, parent)
        # return r

        # return True

    # @functools.singledispatchmethod
    def index(
        self,
        path_or_row: str | int,
        column: int = 0,
        index: QtCore.QModelIndex | None = None,
    ) -> core.ModelIndex:
        if isinstance(path_or_row, int):
            return super().index(path_or_row, column, index)
        return self._iter_path(path_or_row, column, index)

    def _iter_path(self, target: os.PathLike, column: int = 0, parent=None):
        parent = parent or core.ModelIndex()
        target = pathlib.Path(target)
        row_count = self.rowCount(parent)
        for i in range(row_count):
            index = self.index(i, column, parent)
            path = self.filePath(index)
            path = pathlib.Path(path)
            if target == path:
                return index
            elif target.is_relative_to(path):
                if self.canFetchMore(index):
                    self.fetchMore(index)
                return self._iter_path(target, column, index)
        return core.ModelIndex()

    def roleNames(self) -> dict[int, QtCore.QByteArray]:
        return {
            259: QtCore.QByteArray(b"filePermissions"),
            1: QtCore.QByteArray(b"fileIcon"),
            4: QtCore.QByteArray(b"statusTip"),
            3: QtCore.QByteArray(b"toolTip"),
            2: QtCore.QByteArray(b"edit"),
            257: QtCore.QByteArray(b"filePath"),
            0: QtCore.QByteArray(b"display"),
            258: QtCore.QByteArray(b"fileName"),
            5: QtCore.QByteArray(b"whatsThis"),
        }

    def setReadOnly(self, val: bool):
        return NotImplemented

    def isReadOnly(self) -> bool:
        return NotImplemented

    def setOption(self, opt, val):
        return NotImplemented

    def options(self):
        return NotImplemented

    def testOption(self, opt):
        return NotImplemented

    def resolveSymlinks(self, val):
        return NotImplemented

    def setResolveSymlinks(self, val):
        return NotImplemented

    def setNameFilterDisables(self, value):
        return NotImplemented

    def nameFilterDisables(self):
        return NotImplemented

    def setNameFilters(self, filters):
        return NotImplemented

    def nameFilters(self):
        return NotImplemented

    def setFilter(self, filter):
        return NotImplemented

    def filter(self):
        return NotImplemented


if __name__ == "__main__":
    import logging
    import sys

    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    import fsspec

    from prettyqt import widgets

    app = widgets.app()

    fs = fsspec.filesystem("github", org="phil65", repo="prettyqt", path="/")
    print(fs.to_json())
    # root = {"name": "C:/", "size": 0, "type": "directory"}
    root = {"name": "prettyqt", "size": 0, "type": "directory"}
    # print(root)
    model = FSSpecTreeModel(fs, root, False)
    print(model.mimeTypes())
    tree = widgets.TreeView()
    tree.setRootIsDecorated(True)
    tree.setup_dragdrop_move()
    tree.setAlternatingRowColors(True)
    tree.set_model(model)
    tree.set_selection_behaviour("rows")
    idx = model.index(
        "C:\\Users\\phili\\AppData\\Local\\Programs\\Python\\Python310\\Lib"
    )
    print(model.get_permissions(idx))
    print(idx.isValid())
    print(idx.parent().parent().data(constants.DISPLAY_ROLE))
    tree.setUniformRowHeights(True)
    tree.setAnimated(True)
    tree.show()
    app.main_loop()



"""Pdf module."""

from .pdfdocument import PdfDocument
from .pdfbookmarkmodel import PdfBookmarkModel
from .pdfsearchmodel import PdfSearchModel


__all__ = ["PdfDocument", "PdfBookmarkModel", "PdfSearchModel"]



from __future__ import annotations

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtCore, QtWidgets


class ScrollBarMixin(widgets.AbstractSliderMixin):
    value_changed = core.Signal(int)

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation | constants.OrientationStr = "horizontal",
        parent: QtWidgets.QWidget | None = None,
    ):
        if isinstance(orientation, QtCore.Qt.Orientation):
            ori = orientation
        else:
            ori = constants.ORIENTATION[orientation]
        super().__init__(ori, parent)
        self.valueChanged.connect(self.on_value_change)

    def scroll_to_next_row(self):
        """Scroll to the next row."""
        self.setValue(self.value() + self.singleStep())

    def scroll_to_previous_row(self):
        """Scroll to the previous row."""
        self.setValue(self.value() - self.singleStep())


class ScrollBar(ScrollBarMixin, QtWidgets.QScrollBar):
    pass


class SmoothScrollBar(ScrollBar):
    scroll_ended = core.Signal()

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation | constants.OrientationStr = "horizontal",
        parent: QtWidgets.QAbstractScrollArea | None = None,
        animation_duration: int = 500,
        easing: core.easingcurve.TypeStr = "out_cubic",
        trigger: bool = False,
    ):
        super().__init__(orientation, parent)
        self._value = self.value()
        self.widget = parent
        self.animation = core.PropertyAnimation()
        self.animation.apply_to(self.value)
        self.animation.set_easing(easing)
        self.animation.setDuration(animation_duration)
        self.animation.finished.connect(self.scroll_ended)
        self.widget.viewport().installEventFilter(self)
        if trigger:
            self.widget.h_scrollbar.valueChanged.connect(gui.Cursor.fake_mouse_move)

    def mouseMoveEvent(self, e):
        self.animation.stop()
        self._value = self.value()
        super().mouseMoveEvent(e)

    def mousePressEvent(self, e):
        self.animation.stop()
        self._value = self.value()
        super().mousePressEvent(e)

    def mouseReleaseEvent(self, e):
        self.animation.stop()
        self._value = self.value()
        super().mouseReleaseEvent(e)

    def setValue(self, value: int):
        if value == self.value():
            return
        self.animation.stop()
        self.scroll_ended.emit()
        self.animation.set_range(self.value(), value)
        self.animation.start()

    def scroll_by_value(self, value: int):
        """Scroll by given distance."""
        self._value += value
        self._value = min(max(self.minimum(), self._value), self.maximum())
        self.setValue(self._value)

    def scroll_to(self, value: int):
        """Scroll to given position."""
        self._value = value
        self._value = min(max(self.minimum(), self._value), self.maximum())
        self.setValue(self._value)

    def reset_value(self, value):
        self._value = value

    def eventFilter(self, source, event):
        if event.type() == core.Event.Type.Wheel and source == self.widget.viewport():
            self.widget.v_scrollbar.scroll_by_value(-event.angleDelta().y())
            return True
        return False



"""multimedia module.

contains QtMultimedia-based classes
"""

from .audioformat import AudioFormat
from .cameraformat import CameraFormat
from .cameradevice import CameraDevice
from .camera import Camera
from .mediaplayer import MediaPlayer


__all__ = [
    "AudioFormat",
    "Camera",
    "CameraDevice",
    "CameraFormat",
    "MediaPlayer",
]
